import 'package:sqflite/sqflite.dart';

class DatabaseConfig {
  static String path = '';
  static Database? database;

  static dbDelete() {
    deleteDatabase(path);
  }

  static dbClose() {
    database!.close();
    database = null;
  }

  static init({
    String? dbName,
    required List<String> tables,
  }) async {
    final databasesPath = await getDatabasesPath();
    path = databasesPath + (dbName ?? 'sql_query') + '.db';
    database = await openDatabase(path, version: 1, onCreate: (Database db, int version) async {
      for (var i in tables) {
        await db.execute(i);
      }
    });
  }

  static Future<bool> insert({dynamic data, tableName}) async {
    bool success = false;
    await database!.transaction(
      (txn) async {
        try {
          var batch = txn.batch();
          if (data is List) {
            for (Map<String, String> i in data) {
              final Map<String, String> maps = {};
              i.forEach((key, value) {
                final parse = value.replaceAll("'", "|_|_|").replaceAll('"', '|-|-|');
                maps.addAll({key: parse});
              });
              batch.insert(tableName, maps);
            }
          } else {
            batch.insert(tableName, data);
          }
          await batch.commit();
          success = true;
        } catch (_) {
          print(_);
          success = false;
        }
      },
    );
    return success;
  }

  static Future<bool> update({dynamic data, where, tableName}) async {
    bool success = false;
    await database!.transaction(
      (txn) async {
        try {
          var batch = txn.batch();
          if (data is List) {
            for (Map<String, String> i in data) {
              final Map<String, String> maps = {};
              i.forEach((key, value) {
                maps.addAll({key: value.replaceAll("'", "|_|_|").replaceAll('"', '|-|-|')});
              });
              batch.update(tableName, maps, where: where);
            }
          } else {
            batch.update(tableName, data, where: where);
          }
          await batch.commit();
          success = true;
        } catch (_) {
          success = false;
          print(_);
        }
      },
    );
    return success;
  }

  static Future<bool> delete({query}) async {
    bool success = false;
    await database!.transaction(
      (txn) async {
        try {
          var batch = txn.batch();
          batch.rawQuery(query);
          await batch.commit();
          success = true;
        } catch (_) {
          success = false;
          print(_);
        }
      },
    );
    return success;
  }

  static Future<List<Map<String, dynamic>>> get({
    query,
  }) async {
    List<Map<String, dynamic>> list = await database!.rawQuery(
      query,
    );
    List<Map<String, dynamic>> lists = [];
    for (var i in list) {
      Map<String, dynamic> map = {};
      i.forEach((key, value) {
        if (value is String) {
          value = value.replaceAll('|_|_|', "'").replaceAll('|-|-|', '"');
          map.addAll({key: value});
        } else {
          map.addAll({key: value});
        }
      });
      lists.add(map);
    }
    return lists;
  }
}
