import 'package:sql_query/random.dart';

enum DataType { text, integer, double }
enum WhereTypes { and, or, whereIn, orWhereIn, whereBetween, orWhereBetween, whereHas, whereLike, orWhereLike }
enum JoinType { left, cross, inner }

class OrderBy {
  String? query;
  static OrderBy asc(columnName) {
    return OrderBy(query: ' $columnName ASC');
  }

  static OrderBy desc(columnName) {
    return OrderBy(query: ' $columnName DESC');
  }

  static String queryBuilder(List<OrderBy> order) {
    String param = '';
    for (int i = 0; i < order.length; i++) {
      param += (order[i].query ?? '');
      if (i + 1 != order.length && order.length > 1) {
        param += ', ';
      }
    }
    return param;
  }

  OrderBy({this.query});
}

abstract class Joins {
  String? tableName, tableTarget, columnBase, columnName, operator;

  Joins({this.tableName, this.tableTarget, this.columnBase, this.columnName, this.operator});
}

abstract class Wheres {
  String? columnName, operator, value, queryBuilder;
  WhereTypes type;
  List<WhereGroup>? whereGroup;

  Wheres querybuilder(query) {
    queryBuilder = query;
    return this;
  }

  Wheres({required this.columnName, this.operator, this.value, required this.type, this.whereGroup});
}

abstract class WhereGroup {
  String columnName;
  String? operator;
  String? value;
  WhereTypes? type;

  WhereGroup({required this.columnName, this.operator, this.value, this.type});
}

class TableDatabase {
  String tableName;
  List<ColumnDatabase> column;

  TableDatabase({required this.column, required this.tableName});
}

class ColumnDatabase {
  bool primaryKey, autoIncrement, isNull;
  String name;
  DataType typeData;

  ColumnDatabase({this.autoIncrement = false, this.isNull = false, required this.name, required this.typeData, this.primaryKey = false});
}

class OrWhereLike extends Wheres {
  OrWhereLike(String columnName, String? value) : super(columnName: columnName, operator: 'LIKE', value: value, type: WhereTypes.orWhereLike);
}

class WhereLike extends Wheres {
  WhereLike(String columnName, String? value) : super(columnName: columnName, operator: 'LIKE', value: value, type: WhereTypes.whereLike);
}

class Where extends Wheres {
  Where(String columnName, String operator, String? value) : super(columnName: columnName, operator: operator, value: value, type: WhereTypes.and);
}

class OrWhere extends Wheres {
  OrWhere(String columnName, String operator, String? value) : super(columnName: columnName, operator: operator, value: value, type: WhereTypes.or);
}

class WhereIn extends Wheres {
  WhereIn(List<WhereGroup> whereGroup) : super(columnName: null, operator: null, value: null, type: WhereTypes.whereIn, whereGroup: whereGroup);
}

class OrWhereIn extends Wheres {
  OrWhereIn(List<WhereGroup> whereGroup) : super(columnName: null, operator: null, value: null, type: WhereTypes.orWhereIn, whereGroup: whereGroup);
}

class WhereBetween extends Wheres {
  WhereBetween(List<WhereGroup> whereGroup) : super(columnName: null, operator: null, value: null, type: WhereTypes.whereBetween, whereGroup: whereGroup);
}

class OrWhereBetween extends Wheres {
  OrWhereBetween(List<WhereGroup> whereGroup) : super(columnName: null, operator: null, value: null, type: WhereTypes.orWhereBetween, whereGroup: whereGroup);
}

class OrWhereGroup extends WhereGroup {
  OrWhereGroup({required columnName, operator, value}) : super(columnName: columnName, operator: operator, value: value, type: WhereTypes.or);
}

class AndWhereGroup extends WhereGroup {
  AndWhereGroup({required columnName, operator, value}) : super(columnName: columnName, operator: operator, value: value, type: WhereTypes.and);
}

class Join extends Joins {
  Join(column, columnBase, {operator, tableName1, tableName2}) : super(columnBase: columnBase, columnName: column, tableName: tableName1, tableTarget: tableName2, operator: operator);
}

class WhereHas extends Wheres {
  String q = '';
  String? otherTableName;
  String tableName;
  Wheres function(Function(Transaction) query) {
    final transaction = Transaction(where: []);
    query(transaction);

    final select = transaction.select.toString().replaceAll('[', '').replaceAll(']', '');
    q = ' EXISTS ( ';
    q += 'SELECT ${select == 'null' ? '*' : select} FROM $tableName ' + (transaction.joinQuery).replaceAll('{NO_SELECTED_TABLE_NAME}', tableName) + Transaction.query(transaction);
    q += ' ) ';

    return querybuilder(q);
  }

  WhereHas(this.tableName, {this.otherTableName}) : super(columnName: null, type: WhereTypes.whereHas);
}

class Transaction {
  /// List Of [
  /// `OrWhere`,
  /// `Where`,
  /// `WhereIn`,
  /// `orWhereIn`,
  /// `whereBetween`,
  /// `orWhereBetween`,
  /// `WhereHas`
  /// ]
  List<Wheres>? where;
  List<String>? select;
  Map<String, String>? data;
  List<Map<String, String>>? listData;

  /// List of [
  /// `Join`
  /// ]
  List<Joins>? joins;
  int limit;

  String joinQuery = '';

  /// List of [
  /// `OrderBy.asc`,
  /// `OrderBy.desc`
  /// ]
  List<OrderBy>? orderBy = [];
  List<String>? groupBy = [];

  static String query(Transaction data) {
    if (data.where == null) {
      return '';
    }
    String q = '';
    int count = 0;
    for (Wheres i in data.where ?? []) {
      ++count;
      if (count == 1) {
        q += 'WHERE ';
      }
      if (i.type == WhereTypes.whereHas) {
        q += (i.queryBuilder ?? '');
        break;
      }
      if (count > 1 && count <= data.where!.length) {
        if (i.type == WhereTypes.and || i.type == WhereTypes.whereIn || i.type == WhereTypes.whereBetween || i.type == WhereTypes.whereLike) {
          q += ' AND ';
        } else if (i.type == WhereTypes.or || i.type == WhereTypes.orWhereBetween || i.type == WhereTypes.orWhereIn || i.type == WhereTypes.whereLike) {
          q += ' OR ';
        }
      }

      if (i.type == WhereTypes.whereIn || i.type == WhereTypes.orWhereBetween || i.type == WhereTypes.whereBetween || i.type == WhereTypes.orWhereIn) {
        q += ' ( ';
        for (int j = 0; j < i.whereGroup!.length; j++) {
          final tryParse = int.tryParse(i.whereGroup?[j].value ?? '');
          if (i.type == WhereTypes.orWhereBetween || i.type == WhereTypes.whereBetween) {
            if (i.whereGroup!.length != 2) {
              throw ArgumentError('Where Between must have 2 length values');
            }
            if (j < 1) {
              q += (i.whereGroup?[j].columnName ?? '') + ' BETWEEN ';
            }
            if (tryParse == null) {
              final parseDate = DateTime.tryParse(i.whereGroup?[j].value ?? '');
              if (parseDate != null) {
                q += "date(${i.whereGroup?[j].value})";
              } else {
                throw ArgumentError('Where Between Date must (y-m-d)');
              }
            } else {
              q += tryParse.toString();
            }
          } else if (i.type == WhereTypes.whereIn || i.type == WhereTypes.orWhereIn) {
            q += (i.whereGroup?[j].columnName ?? '') + (i.whereGroup?[j].operator ?? '');
            if (tryParse == null) {
              q += '"' + (i.whereGroup![j].value ?? '').replaceAll("'", "|_|_|").replaceAll('"', '|-|-|') + '"';
            } else {
              q += tryParse.toString();
            }
          }

          if ((j + 1) != i.whereGroup!.length && i.whereGroup!.length > 1) {
            if (i.whereGroup?[j].type == WhereTypes.or) {
              q += ' OR ';
            } else if (i.whereGroup?[j].type == WhereTypes.and) {
              q += ' AND ';
            }
          }
        }
        q += ' ) ';
      } else {
        final tryParse = int.tryParse(i.value ?? '');
        q += (i.columnName ?? '') + ' ' + (i.operator ?? '') + ' ';
        if (tryParse == null) {
          q += '"' + (i.value ?? '').replaceAll("'", "|_|_|").replaceAll('"', '|-|-|') + '"';
        } else {
          q += tryParse.toString();
        }
      }
    }
    return q;
  }

  Transaction join(tableName, {tableTargetName, columnName, baseColumnName, Function(Transaction query)? function}) {
    final trans = Transaction(
      select: select,
      data: data,
      limit: limit,
      listData: listData,
      orderBy: orderBy,
      where: where,
      joinQuery: joinQuery,
      groupBy: groupBy,
    );
    String q = ' INNER JOIN $tableName ON ';
    if (function != null) {
      trans.joins ??= [];
      function(trans);

      for (int i = 0; i < trans.joins!.length; i++) {
        q += (trans.joins![i].tableName ?? tableName) + '.' + trans.joins![i].columnName;
        q += '=';
        q += (trans.joins![i].tableTarget ?? '{NO_SELECTED_TABLE_NAME}') + '.' + (trans.joins![i].columnBase ?? '');
        if (i + 1 < trans.joins!.length) {
          q += ' AND ';
        }
      }
    } else {
      q += tableName + '.' + columnName;
      q += '=';
      q += (tableTargetName ?? '{NO_SELECTED_TABLE_NAME}') + '.' + baseColumnName;
    }
    trans.joinQuery += q + ' ';
    return trans;
  }

  Transaction leftJoin(tableName, {tableTargetName, columnName, baseColumnName, Function(Transaction query)? function}) {
    final trans = Transaction(
      select: select,
      data: data,
      limit: limit,
      listData: listData,
      orderBy: orderBy,
      where: where,
      joinQuery: joinQuery,
      groupBy: groupBy,
    );
    String q = ' LEFT JOIN $tableName ON ';
    if (function != null) {
      trans.joins ??= [];
      function(trans);

      for (int i = 0; i < trans.joins!.length; i++) {
        q += (trans.joins![i].tableName ?? tableName) + '.' + trans.joins![i].columnName;
        q += '=';
        q += (trans.joins![i].tableTarget ?? '{NO_SELECTED_TABLE_NAME}') + '.' + (trans.joins![i].columnBase ?? '');
        if (i + 1 < trans.joins!.length) {
          q += ' AND ';
        }
      }
    } else {
      q += tableName + '.' + columnName;
      q += '=';
      q += (tableTargetName ?? '{NO_SELECTED_TABLE_NAME}') + '.' + baseColumnName;
    }

    trans.joinQuery += q + ' ';
    return trans;
  }

  Transaction crossJoin(tableName, {tableTargetName, columnName, baseColumnName, Function(Transaction query)? function}) {
    final trans = Transaction(
      select: select,
      data: data,
      limit: limit,
      listData: listData,
      orderBy: orderBy,
      where: where,
      joinQuery: joinQuery,
      groupBy: groupBy,
    );
    String q = ' CROSS JOIN $tableName ON ';
    if (function != null) {
      trans.joins ??= [];
      function(trans);

      for (int i = 0; i < trans.joins!.length; i++) {
        q += (trans.joins![i].tableName ?? tableName) + '.' + trans.joins![i].columnName;
        q += '=';
        q += (trans.joins![i].tableTarget ?? '{NO_SELECTED_TABLE_NAME}') + '.' + (trans.joins![i].columnBase ?? '');
        if (i + 1 < trans.joins!.length) {
          q += ' AND ';
        }
      }
    } else {
      q += tableName + '.' + columnName;
      q += '=';
      q += (tableTargetName ?? '{NO_SELECTED_TABLE_NAME}') + '.' + baseColumnName;
    }

    trans.joinQuery += q + ' ';
    return trans;
  }

  Transaction selfJoin(baseColumnName, columnName, {Function(Transaction trx, String tableName, String baseTableName)? function, JoinType? join}) {
    final trans = Transaction(
      select: select,
      data: data,
      limit: limit,
      listData: listData,
      orderBy: orderBy,
      where: where,
      joinQuery: joinQuery,
      groupBy: groupBy,
    );
    final rand = getRandomString(9);
    String joinType = 'INNER JOIN';
    if (join == JoinType.left) {
      joinType = 'LEFT JOIN';
    } else if (join == JoinType.cross) {
      joinType = 'CROSS JOIN';
    }
    String q = ' $joinType {NO_SELECTED_TABLE_NAME} $rand ON ';
    q += ('{NO_SELECTED_TABLE_NAME}') + '.' + baseColumnName;
    q += '=';
    q += rand + '.' + columnName;

    if (function != null) {
      /// Select

      function(trans, rand, '{NO_SELECTED_TABLE_NAME}');
    }

    trans.joinQuery += q + ' ';
    return trans;
  }

  Transaction({this.data, this.where, this.listData, this.select, this.limit = 100, this.orderBy, this.joinQuery = '', this.groupBy, this.joins});
}

class DatabaseQuery {
  Transaction? get, update, delete, insert;

  DatabaseQuery({this.delete, this.get, this.insert, this.update});
}
