import '../database_config.dart';
import '../random.dart';

class Query {
  String table;
  String _lastQuery = '';
  String _joinQuery = '';
  String _closeQuery = '';
  String _select = '';
  final List<Map<String, String>> _data = [];

  Query({required this.table});

  Query listData(List<Map<String, String>> listData) {
    _data.insertAll(_data.length, listData);
    return this;
  }

  Query data(Map<String, String> data) {
    _data.add(data);
    return this;
  }

  Query limit(int limit) {
    if (!_joinQuery.contains('LIMIT')) {
      _closeQuery += ' LIMIT $limit ';
    }
    return this;
  }

  /// ``` table1.id DESC , table2.id ASC ```
  Query orderBy(List<String> columnName) {
    final order = columnName.toString().replaceAll('[', '').replaceAll(']', '');
    _closeQuery += ' ORDER BY $order ';
    return this;
  }

  Query groupBy(List<String> columnName) {
    final groupB = columnName.toString().replaceAll('[', '').replaceAll(']', '');
    if (groupB != 'null') {
      _closeQuery += ' GROUP BY $groupB';
    }
    return this;
  }

  Query select(List<String> column) {
    if (_select != '') {
      _select += ', ';
    }
    _select += column.toString().replaceAll('[', '').replaceAll(']', '');
    return this;
  }

  Query where({column, operator, value}) {
    if (!_lastQuery.contains('WHERE')) {
      _lastQuery += ' WHERE ';
    } else {
      _lastQuery += ' AND ';
    }

    _lastQuery += (column) + ' ' + (operator ?? '=') + ' ' + (value ?? 'NULL');
    return this;
  }

  Query orWhere({column, operator, value}) {
    if (!_lastQuery.contains('WHERE')) {
      _lastQuery += ' WHERE ';
    } else {
      _lastQuery += ' OR ';
    }

    _lastQuery += (column) + ' ' + (operator ?? '=') + ' ' + (value ?? 'NULL');
    return this;
  }

  Query whereHas(tableName, Query Function(Query)? function) {
    if (!_lastQuery.contains('WHERE')) {
      _lastQuery += ' WHERE ';
    } else {
      _lastQuery += ' AND ';
    }
    String has = ' EXISTS ( ';
    if (function != null) {
      Query q = function(Query(table: tableName));
      has += 'SELECT ${q._select == '' ? '*' : q._select} FROM $table ' + q._joinQuery + ' ' + q._lastQuery;
    }
    has += ' ) ';
    _lastQuery += has + ' ';
    return this;
  }

  Query orWhereHas(tableName, Query Function(Query)? function) {
    if (!_lastQuery.contains('WHERE')) {
      _lastQuery += ' WHERE ';
    } else {
      _lastQuery += ' OR ';
    }
    String has = ' EXISTS ( ';
    if (function != null) {
      Query q = function(Query(table: tableName));
      has += 'SELECT ${q._select == '' ? '*' : q._select} FROM $table ' + q._joinQuery + ' ' + q._lastQuery;
    }
    has += ' ) ';
    _lastQuery += has + ' ';
    return this;
  }

  Query on(tableColumn, baseColumn) {
    if (_joinQuery.contains('{TABLE_COLUMN_FILL}')) {
      _joinQuery += ' AND ';
    }

    _joinQuery += '{TABLE_COLUMN_FILL}.' + tableColumn + '=' + table + '.' + baseColumn;
    return this;
  }

  Query join(tableName, {tableColumn, baseColumn, baseTableCustom, Query Function(Query)? function}) {
    if (function != null) {
      _joinQuery += 'INNER JOIN  $tableName ON ';

      Query q = function(Query(table: baseTableCustom ?? table));
      _joinQuery += q._joinQuery.replaceAll('{TABLE_COLUMN_FILL}', baseTableCustom ?? table) + ' ' + q._lastQuery.replaceAll('WHERE', 'AND');
    } else {
      _joinQuery += ' INNER JOIN $tableName ON ';
      _joinQuery += tableName + '.' + tableColumn + '=' + (baseTableCustom ?? table) + '.' + baseColumn;
    }
    return this;
  }

  Query leftJoin(tableName, {tableColumn, baseColumn, baseTableCustom, Query Function(Query)? function}) {
    if (function != null) {
      _joinQuery += 'LEFT JOIN  $tableName ON ';

      Query q = function(Query(table: baseTableCustom ?? table));
      _joinQuery += q._joinQuery.replaceAll('{TABLE_COLUMN_FILL}', baseTableCustom ?? table) + ' ' + q._lastQuery.replaceAll('WHERE', 'AND');
    } else {
      _joinQuery += ' LEFT JOIN $tableName ON ';
      _joinQuery += tableName + '.' + tableColumn + '=' + (baseTableCustom ?? table) + '.' + baseColumn;
    }
    return this;
  }

  Query crossJoin(tableName, {tableColumn, baseColumn, baseTableCustom, Query Function(Query)? function}) {
    if (function != null) {
      _joinQuery += ' ( ';
      _joinQuery += 'CROSS JOIN  $tableName ON ';

      Query q = function(Query(table: baseTableCustom ?? table));
      _joinQuery += q._joinQuery.replaceAll('{TABLE_COLUMN_FILL}', baseTableCustom ?? table) + ' ' + q._lastQuery.replaceAll('WHERE', 'AND');
      _joinQuery += ' ) ';
    } else {
      _joinQuery += ' CROSS JOIN $tableName ON ';
      _joinQuery += tableName + '.' + tableColumn + '=' + (baseTableCustom ?? table) + '.' + baseColumn;
    }
    return this;
  }

  Query selfJoin(tableName, {columnName, baseColumnName, Query Function(Query q, String tableName, String baseTableName)? function}) {
    if (function != null) {
      final rand = getRandomString(9);
      _joinQuery += ' ( ';
      _joinQuery += 'INNER JOIN $tableName $rand ON ';

      Query q = function(Query(table: tableName), tableName, rand);
      _joinQuery += q._joinQuery.replaceAll('{TABLE_COLUMN_FILL}', rand) + ' ' + q._lastQuery.replaceAll('WHERE', 'AND');
      _joinQuery += ' ) ';
    } else {
      final rand = getRandomString(9);
      _joinQuery += ' INNER JOIN $tableName $rand ON ';
      _joinQuery += tableName + '.' + baseColumnName;
      _joinQuery += '=';
      _joinQuery += rand + '.' + columnName;
    }
    return this;
  }

  selfLeftJoin(tableName, baseColumnName, columnName, {Query Function(Query q, String tableName, String baseTableName)? function}) {
    if (function != null) {
      final rand = getRandomString(9);
      _joinQuery += 'LEFT JOIN  $rand ON ';

      Query q = function(Query(table: tableName), tableName, rand);
      _joinQuery += q._joinQuery.replaceAll('{TABLE_COLUMN_FILL}', rand) + ' ' + q._lastQuery.replaceAll('WHERE', 'AND');
    } else {
      final rand = getRandomString(9);
      _joinQuery += ' LEFT JOIN $tableName $rand ON ';
      _joinQuery += tableName + '.' + baseColumnName;
      _joinQuery += '=';
      _joinQuery += rand + '.' + columnName;
    }
    return this;
  }

  crossLeftJoin(tableName, baseColumnName, columnName, {Query Function(Query q, String tableName, String baseTableName)? function}) {
    if (function != null) {
      final rand = getRandomString(9);
      _joinQuery += 'CROSS JOIN  $rand ON ';

      Query q = function(Query(table: tableName), tableName, rand);
      _joinQuery += q._joinQuery.replaceAll('{TABLE_COLUMN_FILL}', rand) + ' ' + q._lastQuery.replaceAll('WHERE', 'AND');
    } else {
      final rand = getRandomString(9);
      _joinQuery += ' CROSS JOIN $tableName $rand ON ';
      _joinQuery += tableName + '.' + baseColumnName;
      _joinQuery += '=';
      _joinQuery += rand + '.' + columnName;
    }
    return this;
  }

  Future<List<Map<String, dynamic>>> insert() async {
    List<Map<String, dynamic>> result = [];
    final res = await DatabaseConfig.insert(data: _data, tableName: table);
    if (res) {
      result.add({'status': 'success', 'message': 'data has been Added'});
    }
    return result;
  }

  Future<List<Map<String, dynamic>>> update() async {
    List<Map<String, dynamic>> result = [];
    final res = await DatabaseConfig.update(data: _data, tableName: table, where: _lastQuery);
    if (res) {
      result.add({'status': 'success', 'message': 'data has been Updated'});
    }
    return result;
  }

  Future<List<Map<String, dynamic>>> delete() async {
    List<Map<String, dynamic>> result = [];
    final lastQ = 'DELETE FROM $table ' + _joinQuery + ' ' + _lastQuery + ' ' + _closeQuery;
    final res = await DatabaseConfig.delete(query: lastQ);
    if (res) {
      result.add({'status': 'success', 'message': 'data has been deleted'});
    }
    return result;
  }

  Future<List<Map<String, dynamic>>> get() async {
    final lastQ = 'SELECT ${_select == '' ? '*' : _select} FROM $table ' + _joinQuery + ' ' + _lastQuery + ' ' + _closeQuery;
    return await DatabaseConfig.get(query: lastQ);
  }
}
